"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GitbookDownloader = void 0;
const turndown_1 = __importDefault(require("turndown"));
const pingFetch_1 = require("./pingFetch");
const linkedom_1 = require("linkedom");
/**
 * A very basic logger that wraps console methods.
 */
const logger = {
    info: console.log,
    warn: console.warn,
    error: console.error,
    debug: process.env.DEBUG ? console.debug : () => { },
};
exports.default = logger;
/**
 * GitbookDownloader class for converting Gitbook pages to markdown
 */
class GitbookDownloader {
    /**
     * Creates a new GitbookDownloader instance
     * @param baseUrl - The base URL of the Gitbook documentation
     * @param options - Configuration options
     */
    constructor(baseUrl, options = {}) {
        this.baseUrl = baseUrl;
        // Determine site type if not specified
        let siteType = options.siteType;
        if (!siteType) {
            // Check if URL contains 'docs', 'documentation', 'guide', 'manual', etc.
            const docsKeywords = [
                "docs",
                "documentation",
                "guide",
                "manual",
                "help",
                "reference",
                "api",
            ];
            const urlLower = this.baseUrl.toLowerCase();
            siteType = docsKeywords.some((keyword) => urlLower.includes(keyword))
                ? "docs"
                : "landing";
            logger.info(`Auto-detected page type: ${siteType}`);
        }
        this.options = {
            fetchDelayMs: 500,
            pingDelayMs: 150,
            maxRetries: 3,
            retryDelay: 500,
            headers: {},
            outputPath: "output.md",
            maxConcurrency: 100,
            siteType: siteType,
            ...options,
        };
        this.status = {
            totalPages: 0,
            currentPage: 0,
            currentUrl: "",
            status: "idle",
            pagesScraped: [],
        };
        this.visitedUrls = new Set();
        this.pages = new Map();
        this.contentHash = new Map();
        this.turndownService = new turndown_1.default({
            headingStyle: "atx",
            codeBlockStyle: "fenced",
            hr: "---",
            bulletListMarker: "*",
        });
        // Configure TurndownService
        this.turndownService.addRule("codeBlocks", {
            filter: ["pre", "code"],
            replacement: function (content, node) {
                const element = node;
                const language = element.getAttribute("class")?.replace("language-", "") || "";
                return `\n\`\`\`${language}\n${content}\n\`\`\`\n`;
            },
        });
        // Add rule for tables
        this.turndownService.addRule("tables", {
            filter: ["table"],
            replacement: function (content) {
                return "\n\n" + content + "\n\n";
            },
        });
        // Add custom rule for images to fix [object Object] issue
        this.turndownService.addRule('images', {
            filter: 'img',
            replacement: function (content, node) {
                const element = node;
                const alt = element.getAttribute('alt') || '';
                const title = element.getAttribute('title') || '';
                if (title) {
                    return `(img of ${alt}, "${title}")`;
                }
                else if (alt) {
                    return `(img of ${alt})`;
                }
                return `(img)`;
            }
        });
        this.turndownService.addRule("div", {
            filter: "div",
            replacement: function (content, node) {
                // Check if the div contains only text nodes
                if (node.childNodes.length === 1 && node.firstChild?.nodeType === 3) {
                    return content + "\n\n";
                }
                // For other cases, use default behavior
                return content;
            },
        });
        // Add rule for SVG images
        this.turndownService.addRule('svgHandler', {
            filter: function (node) {
                return node.nodeName === 'svg';
            },
            replacement: function (content, node) {
                // In Edge runtime, SVGElement is not defined, so we use a type check approach
                try {
                    // Try to get a title or description from the SVG
                    const element = node;
                    const title = element.getAttribute('title') || '';
                    const alt = element.getAttribute('alt') || '';
                    // For SVGs, we can't easily convert them, so we'll note their presence
                    if (title) {
                        return `(svg of ${alt}, "${title}")`;
                    }
                    else if (alt) {
                        return `(svg of ${alt})`;
                    }
                    return `(svg)`;
                }
                catch (error) {
                    return `(svg)`;
                }
            }
        });
    }
    /**
     * Downloads and converts Gitbook content to markdown
     */
    async download() {
        try {
            this.status.startTime = Date.now();
            this.status.status = "downloading";
            this.status.currentUrl = this.baseUrl;
            // Try to fetch sitemap first
            logger.info("Starting download process...");
            logger.debug("Fetching sitemap...");
            const sitemapUrls = await this._fetchSitemap();
            let urls = [];
            // Checking sitemapUrls
            if (sitemapUrls.length > 1) {
                logger.info({ count: sitemapUrls.length }, "Found URLs in sitemap");
                urls = sitemapUrls;
                // For landing pages, always include the base URL if it's not already in the sitemap
                if (this.options.siteType === "landing" &&
                    !urls.includes(this.baseUrl)) {
                    logger.info("Adding base URL to sitemap URLs for landing page");
                    urls.unshift(this.baseUrl);
                }
            }
            else if (sitemapUrls.length === 1) {
                const onlyUrl = sitemapUrls[0];
                // If the unique URL contains "sitemap.xml" or matches the base URL,
                // it's considered a non-content URL and we'll fall back to navigation extraction.
                if (onlyUrl.includes(".xml") || onlyUrl === this.baseUrl) {
                    logger.warn("Sitemap returned a single non-content URL. Falling back to navigation extraction");
                }
                else {
                    logger.debug({ count: sitemapUrls.length }, "Found unique content URL in sitemap");
                    urls = sitemapUrls;
                    // For landing pages, always include the base URL if it's not already in the sitemap
                    if (this.options.siteType === "landing" &&
                        !urls.includes(this.baseUrl)) {
                        urls.unshift(this.baseUrl);
                    }
                }
            }
            else {
                logger.warn("No sitemap found, falling back to navigation extraction");
            }
            // Fallback to manual navigation extraction
            if (urls.length <= 1) {
                // Fallback to manual navigation extraction
                const initialContent = await this._fetchPage(this.baseUrl);
                if (!initialContent) {
                    throw new Error("Failed to fetch main page");
                }
                const { document } = (0, linkedom_1.parseHTML)(initialContent);
                urls = this._extractNavLinks(document);
                logger.info({ count: urls.length }, "Found URLs through navigation");
                // Add the main page if not already included
                if (!urls.includes(this.baseUrl)) {
                    urls.unshift(this.baseUrl);
                    logger.debug("Added main page URL to processing list");
                }
            }
            // Starting page processing
            this.status.totalPages = urls.length;
            logger.debug({ totalPages: this.status.totalPages }, "Starting page processing");
            // Process pages in parallel with rate limiting
            const chunks = this._chunkArray(urls, this.options.maxConcurrency);
            logger.debug({ chunks: chunks.length, chunkSize: this.options.maxConcurrency }, "Split URLs into chunks");
            for (const chunk of chunks) {
                logger.debug({ urls: chunk.length }, "Processing chunk of URLs");
                await Promise.all(chunk.map((url) => this._processUrl(url)));
                logger.debug({ delay: this.options.fetchDelayMs }, "Waiting before next chunk");
                await this._sleep(this.options.fetchDelayMs);
            }
            logger.info({ processed: this.pages.size }, "Completed page processing");
            this._sortPages();
            const markdown = await this._generateMarkdown();
            this.status.status = "completed";
            logger.info("Download completed successfully");
            return { pages: this.pages, content: markdown };
        }
        catch (error) {
            this.status.status = "error";
            this.status.error = error.message;
            logger.error({ err: error }, "Download failed");
            throw error;
        }
    }
    _sortPages() {
        logger.debug("sorting pages");
        const sortedPages = Array.from(this.pages.values()).sort((a, b) => {
            const aPath = new URL(a.url).pathname;
            const bPath = new URL(b.url).pathname;
            return aPath.localeCompare(bPath);
        });
        this.pages.clear();
        sortedPages.forEach((page, index) => this.pages.set(index, page));
    }
    /**
     * Attempts to fetch and parse the sitemap
     */
    async _fetchSitemap() {
        const sitemapPaths = [
            "/sitemap.xml",
            "/docs/sitemap.xml",
            "/documentation/sitemap.xml",
            "/sitemap-pages.xml",
            "/sitemap_index.xml",
            "/page-sitemap.xml",
            "/post-sitemap.xml",
            "/docs/sitemap-pages.xml",
            "/documentation/sitemap-pages.xml",
            "/sitemap-posts.xml",
            "/docs/sitemap-posts.xml",
            "/documentation/sitemap-posts.xml",
        ];
        const fetchPromises = sitemapPaths.map(async (path) => {
            try {
                const sitemapUrl = new URL(path, this.baseUrl).toString();
                logger.debug({ url: sitemapUrl }, "Trying sitemap");
                const response = await fetch(sitemapUrl);
                if (!response.ok) {
                    return [];
                }
                const content = await response.text();
                const { document } = (0, linkedom_1.parseHTML)(content);
                const locations = document.querySelectorAll("loc");
                const sitemapUrls = Array.from(locations)
                    .map((loc) => loc.textContent)
                    .filter((url) => Boolean(url && url.startsWith(this.baseUrl)));
                if (sitemapUrls.length > 0) {
                    logger.debug({ count: sitemapUrls.length, sitemap: sitemapUrl }, "Found URLs in sitemap");
                }
                return sitemapUrls;
            }
            catch (error) {
                logger.debug({ path }, "No sitemap found");
                return [];
            }
        });
        const results = await Promise.all(fetchPromises);
        return results.flat();
    }
    /**
     * Processes a single URL
     */
    async _processUrl(url) {
        if (this.visitedUrls.has(url)) {
            logger.debug({ url }, "Skipping already visited URL");
            return;
        }
        logger.debug({ url }, "Processing URL");
        this.visitedUrls.add(url);
        const content = await this._fetchPage(url);
        if (content) {
            await this._processPageContent(content, url);
            this.status.currentPage++;
            logger.debug({ current: this.status.currentPage, total: this.status.totalPages }, "Page processed");
        }
    }
    /**
     * Splits an array into chunks for parallel processing
     */
    _chunkArray(array, size) {
        const chunks = [];
        for (let i = 0; i < array.length; i += size) {
            chunks.push(array.slice(i, i + size));
        }
        return chunks;
    }
    /**
     * Gets the current download status
     */
    getStatus() {
        return { ...this.status };
    }
    /**
     * Fetches a page and handles retries using pingfetch
     */
    async _fetchPage(url) {
        let retries = 0;
        while (retries < this.options.maxRetries) {
            try {
                logger.debug("Fetching %s with pingfetch", url);
                const content = await (0, pingFetch_1.fetchPageWithDelay)(url, this.options);
                logger.debug("Fetched %d bytes from %s", content.length, url);
                return content;
            }
            catch (error) {
                if (error instanceof pingFetch_1.RateLimitError) {
                    logger.warn("Rate limit hit, waiting until %s", error.resetTime.toISOString());
                    await this._sleep(error.resetTime.getTime() - Date.now());
                }
                else {
                    logger.error(`Error fetching ${url}: ${error.name}`);
                }
                retries++;
                if (retries === this.options.maxRetries) {
                    throw error;
                }
                logger.warn(`Retry ${retries}/${this.options.maxRetries} after ${this.options.retryDelay} ms`);
                await this._sleep(this.options.retryDelay);
            }
        }
        return null;
    }
    /**
     * Processes page content and extracts navigation links
     */
    async _processPageContent(content, url) {
        const { document } = (0, linkedom_1.parseHTML)(content);
        // extract all text from divs and log that raw text
        // logDivTextContent(document);
        // Extract title
        let title = "";
        const h1 = document.querySelector("h1");
        if (h1) {
            title = h1.textContent?.trim() || "";
        }
        if (!title) {
            const titleTag = document.querySelector("title");
            if (titleTag) {
                title = titleTag.textContent?.split("|")[0]?.trim() || "";
            }
        }
        if (!title) {
            title =
                new URL(url).pathname.split("/").pop()?.replace(/-/g, " ") ||
                    "Introduction";
        }
        logger.debug("Processing page: %s", title);
        // Get main content based on site type
        let mainContent = null;
        if (this.options.siteType === "docs") {
            // Documentation site content extraction strategy
            mainContent = this._extractDocsContent(document);
        }
        else {
            // Landing page content extraction strategy
            mainContent = this._extractLandingContent(document);
        }
        if (!mainContent) {
            mainContent = document.body;
            logger.debug("Falling back to body element for content");
        }
        logger.debug("Found main content element: %s", mainContent.tagName);
        // Clone the content to avoid modifying the original
        const contentClone = mainContent.cloneNode(true);
        // Clean up content based on site type
        if (this.options.siteType === "docs") {
            this._cleanupDocsContent(contentClone);
        }
        else {
            this._cleanupLandingContent(contentClone);
        }
        logDivTextContent(contentClone);
        // Store page data
        const pageIndex = this.pages.size;
        const content_hash = this._hashString(contentClone.innerHTML);
        // Only store if content is unique
        if (!this.contentHash.has(content_hash)) {
            logger.debug("Storing page %d: %s", pageIndex, title);
            this.pages.set(pageIndex, {
                index: pageIndex,
                title,
                content: contentClone.innerHTML,
                url,
            });
            this.contentHash.set(content_hash, pageIndex);
            this.status.pagesScraped.push(title);
        }
        else {
            logger.debug("Skipping duplicate content for %s", title);
        }
    }
    /**
     * Extract content from documentation sites
     */
    _extractDocsContent(document) {
        // if there is a element with id content-container, use that (mintlify docs)
        let mainContent = document.querySelector('[id*="content-container"], [id*="content-area"]');
        if (!mainContent) {
            mainContent = document.querySelector("main");
        }
        if (!mainContent) {
            mainContent = document.querySelector("article");
        }
        if (!mainContent) {
            mainContent = document.querySelector('[id*="content-container"], [id*="content-area"]');
        }
        if (!mainContent) {
            mainContent = document.querySelector('[class*="markdown"], [class*="content"], [class*="article"], [class*="documentation"]');
        }
        if (!mainContent) {
            // Look for the largest content block
            const contentBlocks = Array.from(document.querySelectorAll("div")).filter((div) => {
                const text = div.textContent || "";
                return text.length > 500 && !div.querySelector("nav, header, footer");
            });
            mainContent = contentBlocks.sort((a, b) => (b.textContent?.length || 0) - (a.textContent?.length || 0))[0];
        }
        return mainContent;
    }
    /**
     * Extract content from landing pages
     */
    _extractLandingContent(document) {
        // For landing pages, we want to be more inclusive and capture more content
        // First try to get the main content area
        let mainContent = document.querySelector("main");
        if (!mainContent) {
            // For landing pages, we might want to include multiple sections
            const sections = document.querySelectorAll("section");
            if (sections.length > 0) {
                // Create a container for all sections
                const container = document.createElement("div");
                sections.forEach((section) => {
                    container.appendChild(section.cloneNode(true));
                });
                logger.debug("Using sections container with", sections.length, "sections");
                return container;
            }
        }
        if (!mainContent) {
            // Try to find content by common landing page classes
            mainContent = document.querySelector('[class*="hero"], [class*="banner"], [class*="landing"], [class*="home"]');
        }
        if (!mainContent) {
            // Look for the body content excluding header and footer
            const body = document.body;
            const container = document.createElement("div");
            // Clone all direct children of body except header and footer
            Array.from(body.children).forEach((child) => {
                if (child.tagName.toLowerCase() !== "header" &&
                    child.tagName.toLowerCase() !== "footer" &&
                    !child.classList.contains("navigation") &&
                    !child.classList.contains("nav") &&
                    !child.classList.contains("menu")) {
                    container.appendChild(child.cloneNode(true));
                }
            });
            if (container.children.length > 0) {
                logger.debug("Using body container with", container.children.length, "children");
                return container;
            }
        }
        // If all else fails, create a container with all divs that have substantial content
        const allContentDivs = Array.from(document.querySelectorAll("div")).filter((div) => {
            const text = div.textContent?.trim() || "";
            return text.length > 100;
        });
        if (allContentDivs.length > 0) {
            const container = document.createElement("div");
            allContentDivs.forEach((div) => {
                container.appendChild(div.cloneNode(true));
            });
            logger.debug("Using fallback container with", allContentDivs.length, "content divs");
            return container;
        }
        // If we still can't find content, return the body as a last resort
        logger.debug("No content found, returning body");
        return document.body;
    }
    /**
     * Clean up documentation content
     */
    _cleanupDocsContent(element) {
        // Remove navigation elements
        const removedNav = element.querySelectorAll('nav, aside, header, footer, [class*="navigation"], [class*="sidebar"], [class*="menu"]');
        logger.debug("Removed %d navigation elements", removedNav.length);
        removedNav.forEach((el) => el.remove());
        // Remove scripts and styles
        const removedScripts = element.querySelectorAll("script, style, iframe, noscript");
        logger.debug("Removed %d script/style elements", removedScripts.length);
        removedScripts.forEach((el) => el.remove());
        // Remove navigation links at bottom
        const removedLinks = Array.from(element.querySelectorAll("a")).filter((link) => {
            const text = link.textContent?.toLowerCase() || "";
            return (text.includes("previous") ||
                text.includes("next") ||
                text.includes("up") ||
                text.includes("down"));
        });
        logger.debug("Removed %d navigation links", removedLinks.length);
        removedLinks.forEach((link) => link.parentElement?.remove());
        // Apply general cleanup
        this._cleanupContent(element);
    }
    /**
     * Clean up landing page content
     */
    _cleanupLandingContent(element) {
        // For landing pages, we want to keep more content but still remove unnecessary elements
        // Remove scripts and styles
        const removedScripts = element.querySelectorAll('script, style, iframe[src*="youtube"], iframe[src*="vimeo"], noscript');
        logger.debug("Removed %d script/style elements", removedScripts.length);
        removedScripts.forEach((el) => el.remove());
        // Remove cookie banners, popups, and modals
        const removedElements = element.querySelectorAll('[class*="cookie"], [class*="popup"], [class*="modal"], [class*="dialog"], [id*="cookie"], [id*="popup"], [id*="modal"]');
        logger.debug("Removed %d cookie/popup/modal elements", removedElements.length);
        removedElements.forEach((el) => el.remove());
        // Apply general cleanup
        this._cleanupContent(element);
    }
    /**
     * Clean up the content by removing empty elements and unnecessary attributes
     */
    _cleanupContent(element) {
        // Remove empty elements
        const emptyElements = element.querySelectorAll("p:empty, div:empty, span:empty");
        emptyElements.forEach((el) => el.remove());
        // Remove unnecessary attributes
        const allElements = element.getElementsByTagName("*");
        for (const el of allElements) {
            const allowedAttrs = ["href", "src", "alt", "title", "class"];
            const attrs = el.attributes;
            for (let i = attrs.length - 1; i >= 0; i--) {
                const attr = attrs[i];
                if (!allowedAttrs.includes(attr.name)) {
                    el.removeAttribute(attr.name);
                }
            }
        }
        // Convert relative URLs to absolute
        const links = element.querySelectorAll("a[href]");
        links.forEach((link) => {
            const href = link.getAttribute("href");
            if (href && !href.startsWith("http")) {
                link.setAttribute("href", new URL(href, this.baseUrl).toString());
            }
        });
        const images = element.querySelectorAll("img[src]");
        images.forEach((img) => {
            const src = img.getAttribute("src");
            if (src && !src.startsWith("http")) {
                img.setAttribute("src", new URL(src, this.baseUrl).toString());
            }
        });
    }
    /**
     * Extracts navigation links from the page
     */
    _extractNavLinks(document) {
        const links = [];
        const processedUrls = new Set();
        // Helper function to process links
        const processLink = (href) => {
            if (href && !href.startsWith("#")) {
                try {
                    const absoluteUrl = new URL(href, this.baseUrl).toString();
                    // Only include links that are part of the same documentation
                    if (absoluteUrl.startsWith(this.baseUrl) &&
                        !processedUrls.has(absoluteUrl)) {
                        logger.debug(`Found navigation link: ${absoluteUrl}`);
                        links.push(absoluteUrl);
                        processedUrls.add(absoluteUrl);
                    }
                }
                catch (e) {
                    console.error(`Invalid URL: ${href}`);
                }
            }
        };
        if (this.options.siteType === "docs") {
            // For documentation sites, focus on navigation elements
            // Find GitBook navigation elements
            const navElements = document.querySelectorAll('nav, aside, [class*="navigation"], [class*="sidebar"], [class*="menu"]');
            logger.debug(`Found ${navElements.length} navigation elements`);
            navElements.forEach((nav) => {
                // Look for lists that typically contain navigation
                const navLists = nav.querySelectorAll("ol, ul");
                logger.debug(`Found ${navLists.length} navigation lists in element`);
                navLists.forEach((navList) => {
                    // Process list items in order
                    const listItems = navList.querySelectorAll("li");
                    logger.debug(`Found ${listItems.length} list items in list`);
                    listItems.forEach((li) => {
                        const link = li.querySelector("a[href]");
                        if (link) {
                            processLink(link.getAttribute("href"));
                        }
                    });
                });
                // Also check direct links in nav elements
                nav.querySelectorAll("a[href]").forEach((link) => {
                    processLink(link.getAttribute("href"));
                });
            });
            // Check for GitBook-specific navigation elements
            document
                .querySelectorAll('[class*="page-list"] a[href], [class*="table-of-contents"] a[href]')
                .forEach((link) => {
                processLink(link.getAttribute("href"));
            });
            // Also check for next/prev navigation links
            document
                .querySelectorAll('a[aria-label], a[rel="prev"], a[rel="next"]')
                .forEach((link) => {
                const ariaLabel = link.getAttribute("aria-label")?.toLowerCase() || "";
                const rel = link.getAttribute("rel");
                if (ariaLabel.includes("next") ||
                    ariaLabel.includes("previous") ||
                    rel === "prev" ||
                    rel === "next") {
                    processLink(link.getAttribute("href"));
                }
            });
        }
        else {
            // For landing pages, be more aggressive in finding links
            // Process all links in the document
            document.querySelectorAll("a[href]").forEach((link) => {
                // For landing pages, we want to be more selective about which links we follow
                const href = link.getAttribute("href");
                const linkText = link.textContent?.toLowerCase() || "";
                // Skip links that are likely not content pages
                const skipKeywords = [
                    "login",
                    "signin",
                    "signup",
                    "register",
                    "account",
                    "cart",
                    "checkout",
                    "twitter",
                    "facebook",
                    "instagram",
                    "linkedin",
                ];
                if (skipKeywords.some((keyword) => linkText.includes(keyword))) {
                    return;
                }
                // Skip links to external domains unless they look like they might be content
                if (href && href.startsWith("http") && !href.startsWith(this.baseUrl)) {
                    // Only include external links if they might be content (blog, article, etc.)
                    const contentKeywords = [
                        "blog",
                        "article",
                        "post",
                        "guide",
                        "tutorial",
                        "docs",
                    ];
                    if (!contentKeywords.some((keyword) => href.includes(keyword))) {
                        return;
                    }
                }
                processLink(href);
            });
            // Look for common landing page sections that might contain important links
            const contentSections = document.querySelectorAll('section, [class*="features"], [class*="products"], [class*="solutions"], [class*="services"]');
            contentSections.forEach((section) => {
                section.querySelectorAll("a[href]").forEach((link) => {
                    processLink(link.getAttribute("href"));
                });
            });
        }
        return links;
    }
    /**
     * Generates the final markdown content
     */
    async _generateMarkdown() {
        const sortedPages = Array.from(this.pages.values()).sort((a, b) => a.index - b.index);
        logger.debug(`Generating markdown for ${sortedPages.length} pages`);
        let markdown = "";
        for (const page of sortedPages) {
            logger.debug(`Converting page: ${page.title}`);
            markdown += `# ${page.title}\n\n`;
            markdown += this.turndownService.turndown(page.content);
            markdown += "\n\n---\n\n";
        }
        return markdown.trim();
    }
    /**
     * Utility method for sleeping
     */
    _sleep(ms) {
        return new Promise((resolve) => setTimeout(resolve, ms));
    }
    /**
     * Utility method for hashing strings
     */
    _hashString(str) {
        let hash = 0;
        for (let i = 0; i < str.length; i++) {
            const char = str.charCodeAt(i);
            hash = (hash << 5) - hash + char;
            hash = hash & hash;
        }
        return hash.toString();
    }
}
exports.GitbookDownloader = GitbookDownloader;
function logDivTextContent(element) {
    const divs = element.querySelectorAll("div");
    let divText = "";
    divs.forEach((div) => {
        divText += div.textContent ? `${div.textContent.trim()}\n` : "";
    });
}
